from cobaya.likelihoods.planck_2018_highl_CamSpec2021.planck_2018_CamSpec2021_python import Planck2018CamSpec2021Python


class TTTEEE(Planck2018CamSpec2021Python):
    r"""
    High-$\ell$ temperature and polarization \textsc{CamSpec} 12.6HMcl likelihood Planck 2018 data
    \cite{efstathiou2019}.
    """
    pass
