from cobaya.likelihoods.bao.desi_2024_bao_all import desi_2024_base


class desi_2024_bao_lrgpluselg_z1(desi_2024_base):
    r"""
    DESI BAO likelihood for LRG+ELG_LOPnotqso in 0.8 < z < 1.1.
    """
